# -*- coding: utf-8 -*-

"""
scikit-ci enables a centralized and simpler CI configuration for Python
extensions.
"""

from .constants import STEPS
from .driver import execute_step
from .exceptions import SKCIError
from ._version import get_versions

__author__ = 'The scikit-build team'
__email__ = 'scikit-build@googlegroups.com'
__version__ = get_versions()['version']
del get_versions

__all__ = ["execute_step", "SKCIError", "STEPS"]




import errno
import os

from contextlib import contextmanager
from functools import wraps


class ContextDecorator(object):
    """A base class or mixin that enables context managers to work as
    decorators."""

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def __enter__(self):
        # Note: Returning self means that in "with ... as x", x will be self
        return self

    def __exit__(self, typ, val, traceback):
        pass

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwds):  # pylint:disable=missing-docstring
            with self:
                return func(*args, **kwds)
        return inner


class push_dir(ContextDecorator):
    """Context manager to change current directory.
    """
    def __init__(self, directory=None, make_directory=False):
        """
        :param directory:
          Path to set as current working directory. If ``None``
          is passed, ``os.getcwd()`` is used instead.

        :param make_directory:
          If True, ``directory`` is created.
        """
        self.directory = None
        self.make_directory = None
        self.old_cwd = None
        super(push_dir, self).__init__(
            directory=directory, make_directory=make_directory)

    def __enter__(self):
        self.old_cwd = os.getcwd()
        if self.directory:
            if self.make_directory:
                mkdir_p(self.directory)
            os.chdir(self.directory)
        return self

    def __exit__(self, typ, val, traceback):
        os.chdir(self.old_cwd)


def mkdir_p(path):
    """Ensure directory ``path`` exists. If needed, parent directories
    are created.

    Adapted from http://stackoverflow.com/a/600612/1539918
    """
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:  # pragma: no cover
            raise


@contextmanager
def push_env(**kwargs):
    """This context manager allow to set/unset environment variables.
    """
    saved_env = dict(os.environ)
    for var, value in kwargs.items():
        if value is not None:
            os.environ[var] = value
        elif var in os.environ:
            del os.environ[var]
    yield
    os.environ.clear()
    for (saved_var, saved_value) in saved_env.items():
        os.environ[saved_var] = saved_value


def captured_lines(cap):
    """Given a ``capsys`` or ``capfd`` pytest fixture, return
     a tuple of the form ``(out_lines, error_lines)``.

    See http://doc.pytest.org/en/latest/capture.html
    """
    out, err = cap.readouterr()
    return (out.rstrip().replace(os.linesep, "\n").split("\n"),
            err.rstrip().replace(os.linesep, "\n").split("\n"))


def display_captured_text(output_lines, error_lines, with_lineno=True):
    """Display the content of captured ``output_lines`` and
    ``error_lines``.

    Here is an example of display::

        [Output]
        0: This is an output line
        1: And an other

        [Error]
        0: This is an error line

    Note that your are responsible to protect this call using
    the ``disabled()`` context manager.

    See http://doc.pytest.org/en/latest/capture.html
    """

    def display_lines(title, lines):
        print("\n[%s]" % title)
        for lineno, line in enumerate(lines):
            print("%s%s" % (str(lineno)+": " if with_lineno else "", line))

    display_lines("Output", output_lines)
    display_lines("Error", error_lines)



# -*- coding: utf-8 -*-

"""
Ci command line tool (enable python -m ci syntax)
"""

import argparse
import ci
import os


class _OptionalStep(argparse.Action):
    """Custom action making the ``step`` positional argument with choices
    optional.

    Setting the ``choices`` attribute will fail with an *invalid choice* error.

    Adapted from http://stackoverflow.com/questions/8526675/python-argparse-optional-append-argument-with-choices/8527629#8527629
    """  # noqa: E501
    def __call__(self, parser, namespace, value, option_string=None):
        if value:
            if value not in ci.STEPS:
                message = ("invalid choice: {0!r} (choose from {1})"
                           .format(value,
                                   ', '.join([repr(action)
                                              for action in
                                              ci.STEPS])))

                raise argparse.ArgumentError(self, message)
            setattr(namespace, self.dest, value)


def main():
    """The main entry point to ``ci.py``.

    This is installed as the script entry point.
    """

    version_str = ("This is scikit-ci version %s, imported from %s\n" %
                   (ci.__version__, os.path.abspath(ci.__file__)))

    parser = argparse.ArgumentParser(description=ci.__doc__)
    parser.add_argument(
        "step", type=str, nargs='?', default=ci.STEPS[-1],
        action=_OptionalStep, metavar='STEP',
        help="name of the step to execute. "
             "Choose from: {}. "
             "If no step is specified, all are executed.".format(", ".join(
                [repr(action) for action in ci.STEPS]))
    )
    parser.add_argument(
        "--force", action="store_true",
        help="always execute the steps"
    )
    parser.add_argument(
        "--without-deps", action="store_false",
        help="do not execute dependent steps", dest='with_dependencies'
    )
    parser.add_argument(
        "--clear-cached-env", action="store_true",
        help="clear cached environment (removes 'env.json' file)"
    )
    parser.add_argument(
        "--version", action="version",
        version=version_str,
        help="display scikit-ci version and import information.")
    args = parser.parse_args()

    try:
        ci.execute_step(
            args.step,
            force=args.force,
            with_dependencies=args.with_dependencies,
            clear_cached_env=args.clear_cached_env
        )
    except ci.SKCIError as exc:
        exit(exc)


if __name__ == '__main__':  # pragma: no cover
    main()




import os
import platform
import pyfiglet
import pytest
import subprocess
import sys
import textwrap

from ruamel.yaml.compat import ordereddict

from . import captured_lines, display_captured_text, push_dir, push_env
from ci.constants import SERVICES, SERVICES_ENV_VAR, STEPS
from ci.driver import Driver, dependent_steps, execute_step
from ci.exceptions import SKCIStepExecutionError
from ci.utils import current_service, current_operating_system

"""Indicate if the system has a Windows command line interpreter"""
HAS_COMSPEC = "COMSPEC" in os.environ

"""Version of the tested scikit-ci schema."""
SCHEMA_VERSION = "0.5.0"


def disable_services(environment=os.environ):
    """Clear environment variables associated will all know services
    """
    for any_service in SERVICES:
        # Clear service variable (e.g APPVEYOR)
        if SERVICES_ENV_VAR[any_service] in environment:
            del environment[SERVICES_ENV_VAR[any_service]]
        # Clear variable associated operating system (e.g TRAVIS_OS_NAME)
        if SERVICES[any_service]:
            if SERVICES[any_service] in environment:
                del environment[SERVICES[any_service]]


def enable_service(service, environment=os.environ, operating_system=None):
    """Ensure ``service`` is enabled.

    For multi-os services, you are expected to set ``operating_system``.

    Note that before enabling ``service``, the environment variables (including
    the one specifying which OS is enabled) for all services are first removed
    from the ``environment``.
    """
    disable_services(environment)
    environment[SERVICES_ENV_VAR[service]] = "true"
    if SERVICES[service]:
        assert operating_system is not None
        environment[SERVICES[service]] = operating_system


@pytest.mark.parametrize("service, exception", [
    ('appveyor', None),
    ('circle', None),
    ('travis', None),
    ('invalid', LookupError)
])
def test_current_service(service, exception):
    with push_env():
        if exception is None:
            operating_system = "linux" if SERVICES[service] else None
            enable_service(service, operating_system=operating_system)
            assert current_service() == service
            assert current_operating_system(service) == operating_system
        else:
            disable_services()
            with pytest.raises(exception):
                current_service()


def scikit_steps(tmpdir, service):
    """Given ``tmpdir`` and ``service``, this generator yields
    ``(step, system, environment)`` for all supported steps.
    """

    # Remove environment variables of the form 'SCIKIT_CI_<STEP>`
    for step in STEPS:
        if 'SCIKIT_CI_%s' % step.upper() in os.environ:
            del os.environ['SCIKIT_CI_%s' % step.upper()]

    # By default, a service is associated with only one "implicit" operating
    # system.
    # Service supporting multiple operating system (e.g travis) should be
    # specified below.
    osenv_per_service = {
        "travis": {"linux": "TRAVIS_OS_NAME", "osx": "TRAVIS_OS_NAME"}
    }

    systems = [None]

    osenv = osenv_per_service.get(service, {})
    if osenv:
        systems = osenv.keys()

    for system in systems:

        # Remove leftover 'env.json'
        env_json = tmpdir.join("env.json")
        if env_json.exists():
            env_json.remove()

        # Set variable like CIRCLE="true" allowing to test for the service
        environment = os.environ
        enable_service(service, environment, system)

        for step in STEPS:

            yield step, system, environment


def _generate_scikit_yml_content(service):
    template_step = (
        r"""
        {what}:

          environment:
            WHAT: {what}
          commands:
            - $<PYTHON> -c "import os; print('%s' % os.environ['WHAT'])"
            {command_0}
            {command_1}
            - $<PYTHON> -c "import sys; print('%s.%s.%s' % sys.version_info[:3])"

          appveyor:
            environment:
              SERVICE: appveyor
            commands:
              - $<PYTHON> -c "import os; print('%s / %s' % (os.environ['WHAT'], os.environ['SERVICE']))"

          circle:
            environment:
              SERVICE: circle
            commands:
              - $<PYTHON> -c "import os; print('%s / %s' % (os.environ['WHAT'], os.environ['SERVICE']))"

          travis:
            linux:
              environment:
                SERVICE: travis-linux
              commands:
                - $<PYTHON> -c "import os; print('%s / %s / %s' % (os.environ['WHAT'], os.environ['SERVICE'], os.environ['TRAVIS_OS_NAME']))"
            osx:
              environment:
                SERVICE: travis-osx
              commands:
                - $<PYTHON> -c "import os; print('%s / %s / %s' % (os.environ['WHAT'], os.environ['SERVICE'], os.environ['TRAVIS_OS_NAME']))"
        """  # noqa: E501
    )

    template = (
        """
        schema_version: "{version}"
        {{}}
        """.format(version=SCHEMA_VERSION)
    )

    if HAS_COMSPEC:
        commands = [
            r"""- $<PYTHON> -c "import os; print('expand:%s' % '$<WHAT>')" """,
            r"""- $<PYTHON> -c "import os; print('expand-2:%s' % '$<WHAT>')" """
            ]
    else:
        commands = [
            r"""- $<PYTHON> -c "import os; print('expand:%s' % \"$<WHAT>\")" """,  # noqa: E501
            r"""- $<PYTHON> -c 'import os; print("expand-2:%s" % "$<WHAT>")' """
            ]

    return textwrap.dedent(template).format(
            "".join(
                [textwrap.dedent(template_step).format(
                    what=step,
                    command_0=commands[0],
                    command_1=commands[1]) for step in STEPS
                 ]
            )
    )


def strip_ascii_art(tuple_of_lines):
    """Return lines without the ascii-art representing the step name.
    """
    if not isinstance(tuple_of_lines, tuple):
        tuple_of_lines = (tuple_of_lines,)

    offset = pyfiglet.FigletFont().height
    return tuple([lines[offset + 1:] for lines in tuple_of_lines])


@pytest.mark.parametrize("service", SERVICES)
def test_driver(service, tmpdir, capfd):

    tmpdir.join('scikit-ci.yml').write(
        _generate_scikit_yml_content(service)
    )

    outputs = []

    with push_env():

        for step, system, environment in scikit_steps(tmpdir, service):

            if "PYTHON" not in environment:
                environment["PYTHON"] = "python"

            with push_dir(str(tmpdir)), push_env(**environment):
                execute_step(step)
                output_lines, error_lines = strip_ascii_art(
                    captured_lines(capfd))

            outputs.append((step, system, output_lines, error_lines))

    with capfd.disabled():
        for step, system, output_lines, error_lines in outputs:

            second_line = "%s / %s" % (step, service)
            if system:
                second_line = "%s-%s / %s" % (second_line, system, system)

            try:
                extra_space = "" if HAS_COMSPEC else " "
                assert output_lines[1] == "%s" % step
                assert output_lines[3] == "expand:%s%s" % (extra_space, step)
                assert output_lines[5] == "expand-2:%s" % (
                    step if HAS_COMSPEC else "$<WHAT>")
                assert output_lines[7] == "%s.%s.%s" % sys.version_info[:3]
                assert output_lines[9] == second_line
            except AssertionError as error:
                context = service + (("-" + system) if system else "")
                print("\n[%s: %s]" % (context, step))
                display_captured_text(output_lines, error_lines)
                if sys.version_info[0] > 2:
                    raise error.with_traceback(sys.exc_info()[2])
                else:
                    raise


def test_dependent_steps():
    step = "before_install"
    expected = []
    assert dependent_steps(step) == expected

    step = "build"
    expected = ['before_install', 'install', 'before_build']
    assert dependent_steps(step) == expected


def test_shell_command(tmpdir, capfd):

    if platform.system().lower() == "windows":
        tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
            r"""
            schema_version: "{version}"
            install:
              commands:
                - FOR %G IN (foo bar) DO python -c "print('var %G')"
            """
        ).format(version=SCHEMA_VERSION))
        service = 'appveyor'
    else:
        tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
            r"""
            schema_version: "{version}"
            install:
              commands:
                - for var in foo bar; do python -c "print('var $var')"; done
                - "for var in oof rab; do python -c \"print('var: $var')\"; done"
            """  # noqa: E501
        ).format(version=SCHEMA_VERSION))
        service = 'circle'

    for step, system, environment in scikit_steps(tmpdir, service):

        with push_dir(str(tmpdir)), push_env(**environment):
            execute_step(step)
            output_lines, _ = strip_ascii_art(captured_lines(capfd))

        if step == 'install':
            if platform.system().lower() == "windows":
                assert output_lines[3] == "var foo"
                assert output_lines[6] == "var bar"
            else:
                assert output_lines[1] == "var foo"
                assert output_lines[2] == "var bar"
                assert output_lines[4] == "var: oof"
                assert output_lines[5] == "var: rab"
        else:
            assert len(output_lines) == 0


@pytest.mark.skipif(platform.system().lower() == "windows",
                    reason="not supported")
def test_multi_line_shell_command(tmpdir, capfd):
    if platform.system().lower() == "windows":
        tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
            r"""
            schema_version: "{version}"
            install:
              commands:
                - |
                  for %G in (foo bar) do ^
                  python -c "print('var %G')"
            """
        ).format(version=SCHEMA_VERSION))
        service = 'appveyor'

    else:
        tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
            r"""
            schema_version: "{version}"
            install:
              commands:
                - |
                  for var in foo bar; do
                    python -c "print('var $var')"
                  done
            """
        ).format(version=SCHEMA_VERSION))
        service = 'circle'

    for step, system, environment in scikit_steps(tmpdir, service):

        with push_dir(str(tmpdir)), push_env(**environment):
            execute_step(step)
            output_lines, _ = strip_ascii_art(captured_lines(capfd))

        if step == 'install':
            assert output_lines[3] == "var foo"
            assert output_lines[4] == "var bar"
        else:
            assert len(output_lines) == 0


def _expand_command_test(command, posix_shell, expected):
    environments = {
        "OTHER": "unused",
        "FO": "foo"
    }
    assert (
        Driver.expand_command(command, environments, posix_shell)
        == expected)


@pytest.mark.parametrize("command, posix_shell, expected", [
    (r"""echo "$<FO>", "$<B>", $<FO>""", False, 'echo "foo" , "$<B>" , foo'),
    (r"""echo '$<FO>', '$<B>', $<FO>""", False, "echo 'foo' , '$<B>' , foo"),
    (r"""echo "$<FO>", "$<B>", $<FO>""", True, 'echo "foo" , "$<B>" , foo'),
    (r"""echo '$<FO>', '$<B>', $<FO>""", True, "echo '$<FO>' , '$<B>' , foo"),
])
def test_expand_command(command, posix_shell, expected):
    _expand_command_test(command, posix_shell, expected)


@pytest.mark.parametrize("command, posix_shell, expected", [
    (r"""echo "$<FO>", \
"$<B>", $<FO>""", True, 'echo "foo" , "$<B>" , foo'),
    (r"""echo '$<FO>', \
'$<B>', $<FO>""", True, "echo '$<FO>' , '$<B>' , foo"),
])
def test_expand_command_with_newline(command, posix_shell, expected):
    _expand_command_test(command, posix_shell, expected)


def test_cli(tmpdir):
    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        install:
          commands:
            - "python -c \"with open('install-done', 'w') as file: file.write('')\""
        """  # noqa: E501
    ).format(version=SCHEMA_VERSION))
    service = 'circle'

    environment = dict(os.environ)
    enable_service(service, environment)

    root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    environment['PYTHONPATH'] = root

    subprocess.check_call(
        "python -m ci install",
        shell=True,
        env=environment,
        stderr=subprocess.STDOUT,
        cwd=str(tmpdir)
    )

    assert tmpdir.join("install-done").exists()


def test_cli_force_and_without_deps(tmpdir):
    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        before_install:
          commands:
            - "python -c \"with open('before_install-done', 'w') as file: file.write('')\""
        install:
          commands:
            - "python -c \"with open('install-done', 'w') as file: file.write('')\""
        """  # noqa: E501
    ).format(version=SCHEMA_VERSION))
    service = 'circle'

    environment = dict(os.environ)
    enable_service(service, environment)

    root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    environment['PYTHONPATH'] = root

    #
    # Execute without --force
    #
    subprocess.check_call(
        "python -m ci install",
        shell=True,
        env=environment,
        stderr=subprocess.STDOUT,
        cwd=str(tmpdir)
    )

    # Check that steps were executed
    assert tmpdir.join("before_install-done").exists()
    assert tmpdir.join("install-done").exists()

    tmpdir.join("before_install-done").remove()
    tmpdir.join("install-done").remove()

    #
    # Execute without --force
    #
    subprocess.check_call(
        "python -m ci install",
        shell=True,
        env=environment,
        stderr=subprocess.STDOUT,
        cwd=str(tmpdir)
    )

    # Check that steps were NOT re-executed
    assert not tmpdir.join("before_install-done").exists()
    assert not tmpdir.join("install-done").exists()

    #
    # Execute with --force
    #
    subprocess.check_call(
        "python -m ci install --force",
        shell=True,
        env=environment,
        stderr=subprocess.STDOUT,
        cwd=str(tmpdir)
    )

    # Check that steps were re-executed
    assert tmpdir.join("before_install-done").exists()
    assert tmpdir.join("install-done").exists()

    tmpdir.join("before_install-done").remove()
    tmpdir.join("install-done").remove()

    #
    # Execute with --force and --without-deps
    #
    subprocess.check_call(
        "python -m ci install --force --without-deps",
        shell=True,
        env=environment,
        stderr=subprocess.STDOUT,
        cwd=str(tmpdir)
    )

    # Check that only specified step was re-executed
    assert not tmpdir.join("before_install-done").exists()
    assert tmpdir.join("install-done").exists()

    tmpdir.join("install-done").remove()
    tmpdir.join("env.json").remove()

    #
    # Execute with --without-deps
    #
    subprocess.check_call(
        "python -m ci install --without-deps",
        shell=True,
        env=environment,
        stderr=subprocess.STDOUT,
        cwd=str(tmpdir)
    )

    # Check that only specified step was executed
    assert not tmpdir.join("before_install-done").exists()
    assert tmpdir.join("install-done").exists()


def test_cli_execute_all_steps(tmpdir):
    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        before_install:
          commands:
            - "python -c \"with open('before_install-done', 'w') as file: file.write('')\""
        install:
          commands:
            - "python -c \"with open('install-done', 'w') as file: file.write('')\""
        """  # noqa: E501
    ).format(version=SCHEMA_VERSION))
    service = 'circle'

    environment = dict(os.environ)
    enable_service(service, environment)

    root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    environment['PYTHONPATH'] = root

    subprocess.check_call(
        "python -m ci",
        shell=True,
        env=environment,
        stderr=subprocess.STDOUT,
        cwd=str(tmpdir)
    )

    assert tmpdir.join("before_install-done").exists()
    assert tmpdir.join("install-done").exists()


def test_not_all_operating_system(tmpdir):
    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        install:
          travis:
            osx:
              environment:
                FOO: BAR
        """  # noqa: E501
    ).format(version=SCHEMA_VERSION))
    service = 'travis'

    environment = dict(os.environ)
    enable_service(service, environment, "linux")

    with push_dir(str(tmpdir)), push_env(**environment):
        execute_step("install")


def test_environment_persist(tmpdir, capfd):
    quote = "" if HAS_COMSPEC else "\""
    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        before_install:
          environment:
            FOO: hello
            BAR: world
            EMPTY: ""
          commands:
            - echo {quote}1 [$<FOO>] [$<BAR>] [$<EMPTY>]{quote}
          circle:
            environment:
              BAR: under world
        install:
          environment:
            BAR: beautiful world
          commands:
            - echo {quote}2 [$<FOO>] [$<BAR>] [$<EMPTY>]{quote}
        """
    ).format(quote=quote, version=SCHEMA_VERSION))
    service = 'circle'

    output_lines = []

    with push_dir(str(tmpdir)), push_env():
        enable_service(service)

        execute_step("before_install")
        output_lines.extend(strip_ascii_art(captured_lines(capfd))[0])

        execute_step("install")
        output_lines.extend(strip_ascii_art(captured_lines(capfd))[0])

    assert output_lines[1] == "1 [hello] [under world] []"
    assert output_lines[3] == "2 [hello] [beautiful world] []"


def test_within_environment_expansion(tmpdir, capfd):
    quote = "" if HAS_COMSPEC else "\""
    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        before_install:
          environment:
            FOO: hello
            BAR: $<WHAT>
            REAL_DIR: $<VERY_DIR>\\real
          commands:
            - echo {quote}[$<FOO> $<BAR> $<STRING>]{quote}
            - echo {quote}[\\the\\thing]{quote}
            - echo {quote}[$<FOO_DIR>\\the\\thing]{quote}
            - echo {quote}[$<FOO_DIR>\\the$<REAL_DIR>\\thing]{quote}
        """
    ).format(quote=quote, version=SCHEMA_VERSION))
    service = 'circle'

    environment = dict(os.environ)
    enable_service(service, environment)

    quote_type = "'" if HAS_COMSPEC else "\""
    backslashes = "\\\\\\\\" if HAS_COMSPEC else "\\"

    environment["WHAT"] = "world"
    environment["STRING"] = "of " + quote_type + "wonders" + quote_type
    environment["FOO_DIR"] = "C:\\path\\to"
    environment["VERY_DIR"] = "\\very"

    with push_dir(str(tmpdir)), push_env(**environment):
        execute_step("before_install")
        output_lines, _ = strip_ascii_art(captured_lines(capfd))

    assert output_lines[1] == "[hello world of " + quote_type + "wonders" + quote_type + "]"  # noqa: E501
    assert output_lines[3] == "[\\the\\thing]".replace("\\", backslashes)
    assert output_lines[5] == "[C:\\path\\to\\the\\thing]".replace("\\", backslashes)  # noqa: E501
    assert output_lines[7] == "[C:\\path\\to\\the\\very\\real\\thing]".replace("\\", backslashes)  # noqa: E501


def test_expand_environment(tmpdir, capfd):
    quote = "" if HAS_COMSPEC else "\""
    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        before_install:
          environment:
            SYMBOLS: b;$<SYMBOLS>
          circle:
            environment:
              SYMBOLS: a;$<SYMBOLS>
          commands:
            - echo {quote}before_install [$<SYMBOLS>]{quote}
        install:
          environment:
            SYMBOLS: 9;$<SYMBOLS>
          circle:
            environment:
              SYMBOLS: 8;$<SYMBOLS>
          commands:
            - echo {quote}install [$<SYMBOLS>]{quote}
        """
    ).format(quote=quote, version=SCHEMA_VERSION))
    service = 'circle'

    output_lines = []

    with push_dir(str(tmpdir)), push_env(SYMBOLS="c;d;e"):
        enable_service(service)

        execute_step("before_install")
        output_lines.extend(strip_ascii_art(captured_lines(capfd))[0])

        execute_step("install")
        output_lines.extend(strip_ascii_art(captured_lines(capfd))[0])

    assert output_lines[1] == "before_install [a;b;c;d;e]"
    assert output_lines[3] == "install [8;9;a;b;c;d;e]"


def recursively_expand_environment_vars_data():

    def case_1():
        """All variables to expand are in ``step_env``
        """
        global_env = ordereddict()

        step_env = ordereddict()
        step_env['C'] = '$<B>'
        step_env['B'] = '$<A>'
        step_env['A'] = 'Hello'

        expected_step_env = ordereddict()
        expected_step_env['C'] = 'Hello'
        expected_step_env['B'] = 'Hello'
        expected_step_env['A'] = 'Hello'

        expected_global_env_size = 3

        return step_env, global_env, expected_step_env, expected_global_env_size

    def case_2():
        """Variable to expand are spread across ``step_env`` and ``global_env``.
        """
        global_env = ordereddict()
        global_env['B'] = '$<A>'

        step_env = ordereddict()
        step_env['C'] = '$<B>'
        step_env['A'] = 'Hello'

        expected_step_env = ordereddict()
        expected_step_env['C'] = 'Hello'
        expected_step_env['A'] = 'Hello'

        expected_global_env_size = 3

        return step_env, global_env, expected_step_env, expected_global_env_size

    def case_3():
        """Variable to expand are spread across ``step_env`` and ``global_env``
        with one of ``step_env`` variable that can not be expanded.
        """
        global_env = ordereddict()
        global_env['D'] = '$<C>'
        global_env['C'] = 'Ciao'
        global_env['G'] = '$<F>'

        step_env = ordereddict()
        step_env['H'] = '$<G>'
        step_env['E'] = '$<D>'
        step_env['B'] = '$<A>'
        step_env['A'] = 'Hello'

        expected_step_env = ordereddict()
        expected_step_env['H'] = '$<F>'
        expected_step_env['E'] = 'Ciao'
        expected_step_env['B'] = 'Hello'
        expected_step_env['A'] = 'Hello'

        expected_global_env_size = 7

        return step_env, global_env, expected_step_env, expected_global_env_size

    def case_4():
        """"""

        global_env = ordereddict()
        global_env["A"] = "0"

        step_env = ordereddict()
        step_env["A"] = "1$<A>"

        expected_step_env = ordereddict()
        expected_step_env['A'] = '10'

        expected_global_env_size = 1

        return step_env, global_env, expected_step_env, expected_global_env_size

    return [
        case_1(),
        case_2(),
        case_3(),
        case_4()
    ]


@pytest.mark.parametrize(
    "step_env, global_env, expected_step_env,expected_global_env_size",
    recursively_expand_environment_vars_data()
)
def test_recursively_expand_environment_vars(
        step_env, global_env, expected_step_env, expected_global_env_size):

    step_env_size = len(step_env)

    Driver.recursively_expand_environment_vars(step_env, global_env)

    assert len(global_env) == expected_global_env_size
    assert len(step_env) == step_env_size

    assert step_env == expected_step_env


def test_ci_name_environment_variable(tmpdir, capfd):
    quote = "" if HAS_COMSPEC else "\""
    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        before_install:
          environment:
            FOO: $<CI_NAME>
          commands:
            - echo {quote}ci_name [$<CI_NAME>] foo [$<FOO>]{quote}
        """
    ).format(quote=quote, version=SCHEMA_VERSION))
    service = 'circle'

    with push_dir(str(tmpdir)), push_env():
        enable_service(service)
        execute_step("before_install")
        output_lines, _ = strip_ascii_art(captured_lines(capfd))

    assert output_lines[1] == "ci_name [%s] foo [%s]" % (service, service)


def test_ci_name_reserved_environment_variable(tmpdir):
    quote = "" if HAS_COMSPEC else "\""
    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        before_install:
          environment:
            CI_NAME: foo
        """
    ).format(quote=quote, version=SCHEMA_VERSION))
    service = 'circle'

    environment = dict(os.environ)
    enable_service(service, environment)

    failed = False
    try:
        with push_dir(str(tmpdir)), push_env(**environment):
            execute_step("before_install")
    except ValueError:
        failed = True

    assert failed


def test_step_ordering_and_dependency(tmpdir):
    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        before_install:
          commands:
            - "python -c \"with open('before_install', 'w') as file: file.write('')\""
        install:
          commands:
            - "python -c \"with open('install', 'w') as file: file.write('')\""
        before_build:
          commands:
            - "python -c \"with open('before_build', 'w') as file: file.write('')\""
        build:
          commands:
            - "python -c \"with open('build', 'w') as file: file.write('')\""
        test:
          commands:
            - "python -c \"exit(1)\""
        after_test:
          commands:
            - "python -c \"with open('after_test', 'w') as file: file.write('')\""
        """  # noqa: E501
    ).format(version=SCHEMA_VERSION))
    service = 'circle'

    environment = dict(os.environ)
    enable_service(service, environment)

    with push_dir(str(tmpdir)), push_env(**environment):

        execute_step("install")

        #
        # Check that steps `before_install` and `install` were executed
        #
        env = Driver.read_env()
        assert env['SCIKIT_CI_BEFORE_INSTALL'] == '1'
        assert env['SCIKIT_CI_INSTALL'] == '1'
        assert tmpdir.join('before_install').exists()
        assert tmpdir.join('install').exists()

        # Remove files - This will to make sure the steps are not re-executed
        tmpdir.join('before_install').remove()
        tmpdir.join('install').remove()

        # Check files have been removed
        assert not tmpdir.join('before_install').exists()
        assert not tmpdir.join('install').exists()

        execute_step("build")

        #
        # Check that only `before_build` and `build` steps were executed
        #
        env = Driver.read_env()
        assert env['SCIKIT_CI_BEFORE_INSTALL'] == '1'
        assert env['SCIKIT_CI_INSTALL'] == '1'
        assert env['SCIKIT_CI_BEFORE_BUILD'] == '1'
        assert env['SCIKIT_CI_BUILD'] == '1'

        assert not tmpdir.join('before_install').exists()
        assert not tmpdir.join('install').exists()
        assert tmpdir.join('before_build').exists()
        assert tmpdir.join('build').exists()

        # Remove files - This will to make sure the steps are not re-executed
        tmpdir.join('before_build').remove()
        tmpdir.join('build').remove()

        failed = False
        try:
            execute_step("after_test")
        except SKCIStepExecutionError as exc:
            failed = "exit(1)" in exc.cmd

        #
        # Check that `after_test` step was NOT executed. It should not be
        # execute because `test` step is failing.
        #
        assert failed
        env = Driver.read_env()
        assert env['SCIKIT_CI_BEFORE_INSTALL'] == '1'
        assert env['SCIKIT_CI_INSTALL'] == '1'
        assert env['SCIKIT_CI_BEFORE_BUILD'] == '1'
        assert env['SCIKIT_CI_BUILD'] == '1'
        assert 'SCIKIT_CI_TEST' not in env
        assert 'SCIKIT_CI_AFTER_TEST' not in env

        assert not tmpdir.join('before_install').exists()
        assert not tmpdir.join('install').exists()
        assert not tmpdir.join('before_install').exists()
        assert not tmpdir.join('install').exists()
        assert not tmpdir.join('test').exists()
        assert not tmpdir.join('after_test').exists()

        #
        # Check `force=True` works as expected
        #
        execute_step("install", force=True)

        # Check that steps `before_install` and `install` were re-executed
        env = Driver.read_env()
        assert env['SCIKIT_CI_BEFORE_INSTALL'] == '1'
        assert env['SCIKIT_CI_INSTALL'] == '1'
        assert tmpdir.join('before_install').exists()
        assert tmpdir.join('install').exists()

        tmpdir.join('before_install').remove()
        tmpdir.join('install').remove()

        #
        # Check `force=True` and `with_dependencies=True` work as expected
        #
        execute_step("install", force=True, with_dependencies=False)

        # Check that only step `install` was re-executed
        env = Driver.read_env()
        assert env['SCIKIT_CI_BEFORE_INSTALL'] == '1'
        assert env['SCIKIT_CI_INSTALL'] == '1'
        assert not tmpdir.join('before_install').exists()
        assert tmpdir.join('install').exists()

        tmpdir.join('install').remove()

        #
        # Check `force=False` and `with_dependencies=True` work as expected
        #
        tmpdir.join('env.json').remove()
        execute_step("install", with_dependencies=False)

        # Check that only step `install` was executed
        env = Driver.read_env()
        assert 'SCIKIT_CI_BEFORE_INSTALL' not in env
        assert env['SCIKIT_CI_INSTALL'] == '1'
        assert not tmpdir.join('before_install').exists()
        assert tmpdir.join('install').exists()


def test_clear_env(tmpdir):
    """This test checks that the 'env.json' file is removed when the force
    option is specified.
    """

    tmpdir.join('scikit-ci.yml').write(textwrap.dedent(
        r"""
        schema_version: "{version}"
        """
    ).format(version=SCHEMA_VERSION))

    service = 'circle'

    environment = dict(os.environ)
    enable_service(service, environment)

    with push_dir(str(tmpdir)), push_env(**environment):

        execute_step("test")

        env = Driver.read_env()
        assert env['SCIKIT_CI_TEST'] == '1'

        # Add variable to check env file is effectively removed
        assert 'CLEAR_ENV_TEST' not in env
        env['CLEAR_ENV_TEST'] = '1'
        Driver.save_env(env)
        assert 'CLEAR_ENV_TEST' in Driver.read_env()

        #
        # Re-execute 'test' step clearing environment
        #
        execute_step("test", clear_cached_env=True)

        env = Driver.read_env()
        assert env['SCIKIT_CI_TEST'] == '1'
        assert 'CLEAR_ENV_TEST' not in env



"""
This module defines exceptions commonly used in scikit-ci.
"""

import os
import textwrap


class SKCIError(RuntimeError):
    """Exception raised when an scikit-ci error occurs.
    """
    pass


class SKCIStepExecutionError(SKCIError):
    """Exception raised when an error occurs while executing a step.
    """
    def __init__(self, step, return_code, cmd, output=None):
        self.step = step
        self.return_code = return_code
        self.cmd = cmd
        self.output = output

    def __str__(self):
        return textwrap.dedent(
            r"""
            A command failed while executing {step} step.
              Return code:
                {return_code}
              Command:
                {cmd}
              Working directory:
                {cwd}

            Please see above for more information.
            """.format(
                step=self.step.upper(),
                return_code=self.return_code,
                cmd=self.cmd,
                cwd=os.getcwd()
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import versioneer

from setuptools import setup

with open('README.rst', 'r') as fp:
    readme = fp.read()

with open('HISTORY.rst', 'r') as fp:
    history = fp.read().replace('.. :changelog:', '')

with open('requirements.txt', 'r') as fp:
    requirements = list(filter(bool, (line.strip() for line in fp)))

with open('requirements-dev.txt', 'r') as fp:
    dev_requirements = list(filter(bool, (line.strip() for line in fp)))

# Require pytest-runner only when running tests
pytest_runner = (['pytest-runner>=2.0,<3dev']
                 if any(arg in sys.argv for arg in ('pytest', 'test'))
                 else [])

setup_requires = pytest_runner

setup(
    name='scikit-ci',

    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),

    author='The scikit-build team',
    author_email='scikit-build@googlegroups.com',

    url='https://github.com/scikit-build/scikit-ci',

    description='scikit-ci enables a centralized and simpler CI configuration '
                'for Python extensions.',
    long_description=readme + '\n\n' + history,

    entry_points={'console_scripts': ['ci=ci.__main__:main']},

    packages=['ci'],
    package_data={},
    include_package_data=True,
    zip_safe=False,

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Build Tools'
    ],

    license="Apache",

    keywords='CI Appveyor CircleCI Travis',

    setup_requires=setup_requires,
    install_requires=requirements,
    tests_require=dev_requirements,
)



# -*- coding: utf-8 -*-

"""This module provides an interface to parse and exectute commands found in
``scikit-ci.yml``."""

import errno
import json
import os
import os.path
import re
import ruamel.yaml
import shlex
import subprocess
import sys

from pyfiglet import Figlet

from . import exceptions, utils
from .constants import SCIKIT_CI_CONFIG, SERVICES, STEPS


class DriverContext(object):
    def __init__(self, driver, env_file="env.json"):
        self.driver = driver
        self.env_file = env_file

    def __enter__(self):
        self.driver.load_env(self.env_file)
        return self.driver

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None and exc_value is None and traceback is None:
            self.driver.save_env(self.driver.env, self.env_file)

        self.driver.unload_env()


class Driver(object):
    def __init__(self):
        self.env = None
        self._env_file = None

    @staticmethod
    def log(*s):
        print(" ".join(s))
        sys.stdout.flush()

    @staticmethod
    def read_env(env_file="env.json"):
        if not os.path.exists(env_file):
            return {}
        with open(env_file) as _file:
            return json.load(_file)

    def load_env(self, env_file="env.json"):
        if self.env is not None:
            self.unload_env()

        self.env = {}
        self.env.update(os.environ)
        self._env_file = env_file

        if os.path.exists(self._env_file):
            self.env.update(self.read_env(self._env_file))

        self.env = {str(k): str(v) for k, v in self.env.items()}

    @staticmethod
    def save_env(env, env_file="env.json"):
        with open(env_file, "w") as _file:
            json.dump(env, _file, indent=4)

    def unload_env(self):
        self.env = None

    def check_call(self, *args, **kwds):
        kwds["env"] = kwds.get("env", self.env)
        kwds["shell"] = True

        if "COMSPEC" in os.environ:
            cmd = "cmd.exe /E:ON /V:ON /C \"%s\"" % args[0]
            args = [cmd]

        self.log("[scikit-ci] Executing: %s" % args[0])
        return subprocess.check_call(*args, **kwds)

    def env_context(self, env_file="env.json"):
        return DriverContext(self, env_file)

    @staticmethod
    def expand_environment_vars(text, environment):
        """Return an updated ``text`` string where all occurrences of
        ``$<EnvironmentVarName>`` found in ``environment`` are replaced.
        """
        for name, value in environment.items():
            text = text.replace(
                "$<%s>" % name,
                value.replace("\\", "\\\\").replace("\"", "\\\""))
        return text

    ENV_VAR_REGEX = re.compile(r"\$<[\w\d][\w\d_]*>", re.IGNORECASE)
    """Regular expression matching legal environment variable of the
    form ``$<EnvironmentVarName>``"""

    @staticmethod
    def recursively_expand_environment_vars(step_env, global_env=None):
        """This function will recursively expand all occurrences of
        ``$<EnvironmentVarName>`` found in ``step_env`` and ``global_env``
        values.
        """

        if global_env is None:
            global_env = step_env

        # Keep track of variables that still need to be expanded
        to_be_expanded = set()

        def _expand(names, _work_env, _global_env=None):
            if _global_env is None:
                _global_env = _work_env
            for env_var_name in names:
                # Get the value
                env_var_value = _work_env[env_var_name]

                # Attempt to expand the value
                _work_env[env_var_name] = Driver.expand_environment_vars(
                    env_var_value, _global_env)

                # Keep track of variable names to expand
                if re.match(Driver.ENV_VAR_REGEX, _work_env[env_var_name]):
                    to_be_expanded.add(env_var_name)
                elif env_var_name in to_be_expanded:
                    to_be_expanded.remove(env_var_name)

        # Expand step env values referencing global env variables
        _expand(step_env.keys(), step_env, global_env)
        global_env.update(step_env)

        # Expand variables
        _expand(step_env.keys(), global_env)

        # Expand remaining variables if any
        to_be_expanded_count = len(to_be_expanded)
        while to_be_expanded:
            _expand(to_be_expanded, global_env)
            if to_be_expanded_count == len(to_be_expanded):
                break
            to_be_expanded_count = len(to_be_expanded)

        # Update step environment
        for name in step_env:
            step_env[name] = global_env[name]

    @staticmethod
    def expand_command(command, environments, posix_shell=True):
        """Return an updated ``command`` string where all occurrences of
        ``$<EnvironmentVarName>`` (with a corresponding env variable set) have
        been replaced.

        If ``posix_shell`` is True, only occurrences of
        ``$<EnvironmentVarName>`` in string starting with double quotes will
        be replaced.

        See
        https://www.gnu.org/software/bash/manual/html_node/Double-Quotes.html
        and
        https://www.gnu.org/software/bash/manual/html_node/Single-Quotes.html
        """
        # Strip line continuation characters. There are not required
        # successfully evaluate the expression and were confusion shlex.
        if posix_shell:
            command = command.replace("\\\n", "")

        tokenizer = shlex.shlex(command, posix=False)
        tokenizer.whitespace_split = True
        expanded_lines = []
        expanded_tokens = []
        lineno = 1
        for token in tokenizer:
            expand = not (posix_shell and token[0] == "'" and token[-1] == "'")
            if expand:
                token = Driver.expand_environment_vars(token, environments)

            if tokenizer.lineno > lineno:
                expanded_lines.append(" ".join(expanded_tokens))
                expanded_tokens = []

            expanded_tokens.append(token)
            lineno = tokenizer.lineno

        if expanded_tokens:
            expanded_lines.append(" ".join(expanded_tokens))

        return "\n".join(expanded_lines)

    @staticmethod
    def _raise_if_setting_ci_name(environment):
        # Check if reserved variable are used
        if "CI_NAME" in environment:
            raise ValueError("CI_NAME environment variable can not be set. "
                             "It is reserved to store the name of the current "
                             "CI service (e.g appveyor, circle or travis.")

    @staticmethod
    def parse_config(config_file, stage_name, service_name, global_env):
        with open(config_file) as input_stream:
            data = ruamel.yaml.load(input_stream, ruamel.yaml.RoundTripLoader)

        commands = []
        environment = {}

        if stage_name in data:
            stage = data[stage_name]

            # common to all services
            environment = stage.get("environment", {})
            commands = stage.get("commands", [])

            # Sanity checks
            Driver._raise_if_setting_ci_name(environment)

            # Expand all occurrences of ``$<EnvironmentVarName>``.
            Driver.recursively_expand_environment_vars(environment, global_env)

            if service_name in stage:
                system = stage[service_name]

                # consider service offering multiple operating system support
                if SERVICES[service_name]:
                    operating_system = global_env[SERVICES[service_name]]
                    system = system.get(operating_system, {})

                # if any, get service specific environment
                system_environment = system.get("environment", {})

                # Sanity checks
                Driver._raise_if_setting_ci_name(environment)

                # Expand system environment values
                Driver.recursively_expand_environment_vars(
                    system_environment, global_env)

                # Merge system environment variable back into environment
                environment.update(system_environment)

                # ... and append commands
                commands += system.get("commands", [])

        return environment, commands

    def execute_commands(self, stage_name):

        print(Figlet().renderText(stage_name.upper()))

        service_name = utils.current_service()

        self.env["CI_NAME"] = service_name

        environment, commands = self.parse_config(
            SCIKIT_CI_CONFIG, stage_name, service_name, self.env)

        # Unescape environment variables
        for name in environment:
            value = self.env[name]
            for old, new in [("\\\\", "\\")]:
                value = value.replace(old, new)
            self.env[name] = value

        posix_shell = "COMSPEC" not in os.environ

        for cmd in commands:
            # Expand environment variables used within commands
            cmd = self.expand_command(
                cmd, self.env, posix_shell=posix_shell)
            try:
                self.check_call(cmd.replace("\\\\", "\\\\\\\\"), env=self.env)
            except subprocess.CalledProcessError as exc:
                raise exceptions.SKCIStepExecutionError(
                    stage_name, exc.returncode, exc.cmd, exc.output
                )


def dependent_steps(step):
    if step not in STEPS:  # pragma: no cover
        raise KeyError("invalid step: {}".format(step))
    step_index = STEPS.index(step)
    if step_index == 0:
        return []
    return STEPS[0:step_index]


def execute_step(
        step, force=False, with_dependencies=True, clear_cached_env=False):

    if not os.path.exists(SCIKIT_CI_CONFIG):  # pragma: no cover
        raise OSError(errno.ENOENT, "Couldn't find %s" % SCIKIT_CI_CONFIG)

    if step not in STEPS:  # pragma: no cover
        raise KeyError("invalid step: {}".format(step))

    if clear_cached_env and os.path.exists('env.json'):
        os.remove('env.json')

    depends = dependent_steps(step)

    # If forcing execution, remove SCIKIT_CI_<step> env. variables
    if force:
        env = Driver.read_env()
        steps = [step]
        if with_dependencies:
            steps += depends
        for _step in steps:
            if 'SCIKIT_CI_%s' % _step.upper() in env:
                del env['SCIKIT_CI_%s' % _step.upper()]
        Driver.save_env(env)

    # Skip step if it has already been executed
    if 'SCIKIT_CI_%s' % step.upper() in Driver.read_env():
        return

    # Recursively execute dependent steps
    if with_dependencies and depends:
        execute_step(depends[-1], with_dependencies=with_dependencies)

    d = Driver()
    with d.env_context():
        d.execute_commands(step)
        d.env['SCIKIT_CI_%s' % step.upper()] = '1'
